import numpy as np
import scipy 
from scipy.fftpack import fft2,ifft2,dst,idst
import pandas as pd
from views_transformation_library import utilities

def get_fourier_lag(df,dimensionality,use_stride_tricks=True):

    '''
    get_fourier_lag
    
    Driver function for obtaining spatial lags by using Fourier transforms to perform a 
    fast solution of Poisson's equation
    
    Arguments:
    
    df: dataframe containing one or more features to be lagged
    
    dimensionality: dimensionality of Fourier transforms to be used, which controls the 
    distance-weighting of the spatial lag. This can be 
    
    - 2, resulting in a lagging proportional to -ln(d)
    - 3, resulting in a lagging proportional to 1/d
    
    use_stride_tricks: boolean, controlling whether stride tricks is used for tensor
    transforms
    
    '''

    if dimensionality==2:
        transformer=fft_2D
    
    elif dimensionality==3:
        transformer=fft_3D
    
    else:
         print('Dimensionality ',dimensionality,' not supported.')
         return None
         
    pgids,pgid_to_longlat,longlat_to_pgid,pgid_to_index,\
    index_to_pgid,ncells,power=utilities._map_pgids_2d(df)     
    
    times,time_to_index,index_to_time=utilities._map_times(df)
    
    features=utilities._map_features(df)
    
    tensor4d=utilities._build_4d_tensor(
                                        df,
                                        pgids,
                                        pgid_to_index,
                                        times,
                                        time_to_index,
                                        ncells,
                                        ncells,
                                        pgid_to_longlat,
                                        features,
                                        use_stride_tricks
                                        )
    
    flags=transformer(tensor4d,times,features,time_to_index)
    
    df_flags=flags_to_df(
                         df,
                         flags,
                         times,
                         time_to_index,
                         pgids,
                         features,
                         ncells,
                         ncells,
                         longlat_to_pgid
                         )
    
    return df_flags
    
    
def fft_2D(tensor4d,times,features,time_to_index):

    '''
    fft_2D
    
    Function which uses a 2D Fourier transform to solve Poisson's equation in 2 spatial
    dimensions.
    
    This results in distance weighting of the lagged variable of -ln(d)
    
    
    '''

    # small value to use instead of zero for zero spatial frequencies

    SMALL=1e-3

    dim0=np.shape(tensor4d)[0]
    transformed=np.empty_like(tensor4d)
    
    freqsj=freqsk=None
    rangej=rangek=dim0
    
    for ifeature,feature in enumerate(features):
        for time in times:

            itime=time_to_index[time]
            
            rhomax=np.max(tensor4d[:,:,itime,ifeature])

            # perform 2D FFT on time slice from tensor

            rhojkhat=np.fft.fft2(tensor4d[:,:,itime,ifeature],norm='ortho')

            if freqsj is None:

               # get spatial frequencies

               freqsj=2.*np.pi*np.fft.fftfreq(rhojkhat.shape[0])
               freqsk=2.*np.pi*np.fft.fftfreq(rhojkhat.shape[1])
       
               freqsj[abs(freqsj)<SMALL]=SMALL
               freqsk[abs(freqsk)<SMALL]=SMALL

               # make tensor with values 1/(j**2+k**2)

               divisor=np.ones((rangej,rangek))
       
               for ij in range(rangej):
                   ifreqj=freqsj[ij]
                   ifreqj2=ifreqj*ifreqj
                   for ik in range(rangek):
                       ifreqk=freqsk[ik]
                       ifreqk2=ifreqk*ifreqk
                       divisor[ij,ik]/=(ifreqj2+ifreqk2)

            # generate solution to Poisson's equation on slice

            phijkhat=np.multiply(rhojkhat,divisor)
            
            # take real part of inverse transform
            
            phijk=np.real(np.fft.ifft2(phijkhat,norm='ortho'))
            
            # subtract zero level
            
            phijk-=np.min(phijk)
            
            # rescale lagged variable to have same scale as original feature
            
            phimax=np.max(phijk)

            phimax=np.max([phimax,1.0])
            
            transformed[:,:,itime,ifeature]=phijk*rhomax/phimax

    return transformed


def fft_3D(tensor4d,times,features,time_to_index):

    '''
    fft_3D
    
    Function which uses a 3D Fourier transform to solve Poisson's equation in 3 spatial
    dimensions.
    
    This results in distance weighting of the lagged variable of 1/d
    
    This is done by embedding the flat array from the original feature as the midplane of 
    a cube, so that the problem effectively becomes 3-dimensional.
    
    To save memory and time, the number of Fourier modes used in the extra dimension is 
    reduced by a factor LDIVISOR. Experimentation shows that adequate resolution can be
    achieved with LDIVISOR set to 4
    
    
    '''

    # factor by which to reduce number of Fourier modes in extra dimension

    LDIVISOR=4
    
    # small value to use instead of zero for zero spatial frequencies
    
    SMALL=1e-3
    
    dim0=np.shape(tensor4d)[0]
    
    transformed=np.empty_like(tensor4d)

    freqsj=freqsk=freqsl=None
    rangej=rangek=dim0
    rangel=int(dim0/LDIVISOR)
    
    for ifeature,feature in enumerate(features):
        for time in times:
          
            itime=time_to_index[time]
            
            rhomax=np.max(tensor4d[:,:,itime,ifeature])
            
            # embed time slice from tensor into a 3D tensor

            rhojkl=np.zeros((rangej,rangek,rangel))
        
            icentrej=icentrek=int(rangej/2)
            icentrel=int(rangel/2)
        
            rhojkl[:,:,icentrel]=tensor4d[:,:,itime,ifeature]
            
            # perform N-dimensional FFT on embedded time slice

            rhojklhat=np.fft.fftn(rhojkl,norm='ortho')

            if freqsj is None:

               # get spatial frequencies

               freqsj=2.*np.pi*np.fft.fftfreq(rhojklhat.shape[0])
               freqsk=2.*np.pi*np.fft.fftfreq(rhojklhat.shape[1])
               freqsl=2.*np.pi*np.fft.fftfreq(rhojklhat.shape[2])
       
               freqsj[abs(freqsj)<SMALL]=SMALL
               freqsk[abs(freqsk)<SMALL]=SMALL
               freqsl[abs(freqsl)<SMALL]=SMALL

               # make tensor with values 1/(j**2+k**2+l**2)

               divisor=np.ones((rangej,rangek,rangel))
       
               for ij in range(rangej):
                   ifreqj=freqsj[ij]
                   ifreqj2=ifreqj*ifreqj
                   for ik in range(rangek):
                       ifreqk=freqsk[ik]
                       ifreqk2=ifreqk*ifreqk
                       for il in range(rangel):
                           ifreql=freqsl[il]
                           ifreql2=ifreql*ifreql
            
                           divisor[ij,ik,il]/=(ifreqj2+ifreqk2+ifreql2)

            # generate solution to Poisson's equation on embedded slice

            phijklhat=np.multiply(rhojklhat,divisor)
            
            # take real part of inverse transform
            
            phijkl=np.real(np.fft.ifftn(phijklhat,norm='ortho'))
            
            # subtract zero level
            
            phijkl-=np.min(phijkl)
            
            # rescale lagged variable to have same scale as original feature
            
            phimax=np.max(phijkl)

            phimax=np.max([phimax,1.0])

            transformed[:,:,itime,ifeature]=phijkl[:,:,icentrel]*rhomax/phimax
        
    return transformed

def flags_to_df(
    df,
    flags,
    times,
    time_to_index,
    pgids,
    features,
    longrange,
    latrange,
    longlat_to_pgid
):

    final=np.empty((len(times)*len(pgids),len(features)))

    # convert 4d spatial lag tensor to flat table

    ipgid=0
    npg=len(pgids)
    pgids_for_index=[]

    for ilong in range(1,longrange+2):
        for ilat in range(1,latrange+2):
            try:
                pgid=longlat_to_pgid[(ilong,ilat)]
                pgids_for_index.append(pgid)
                for time in times:
                    tindex=time_to_index[time]
                    indx=tindex*npg+ipgid
                    final[indx,:]=flags[ilong,ilat,tindex,:]

                ipgid+=1
            except:
                pass
            
    index_names=df.index.names
        
    df_index=pd.MultiIndex.from_product([list(times),pgids_for_index],names=index_names)
            
    colnames=['flag_'+feature for feature in features]
    df_flags=pd.DataFrame(data=final,columns=colnames,index=df_index)

    df_flags=df_flags.sort_index()

    return df_flags
    





import numpy as np
import pandas as pd

from scipy import ndimage

from views_transformation_library import utilities
   
def get_splag4d(
    df,
    use_stride_tricks=True,
    kernel_inner=1,kernel_width=1,kernel_power=0,norm_kernel=0

):

    '''
    splag4d created 19/03/2021 by Jim Dale
    
    Performs spatial lags on a dataframe by transforming from flat format to 4d tensor 
    with dimensions longitude x latitude x time x features.
    
    Spatial lagging can then be done as a 2d convolution on long-lat slices using
    scipy convolution algorithms.
    
    Arguments:
    
    df:                a dataframe of series to be splagged
    
    use_stride_tricks: boolean, decide to use stride_tricks or not (optional, 
                       defaults to True)
   
    kernel_inner:      inner border of convolution region (set to 1 to exclude central
                       cell)

    kernel_width:      width in cells of kernel, so outer radius of kernel =
                       kernel_inner + kernel_width

    kernel_power:      weight values of cells by (distance from centre of kernel)**
                       (-kernel_power) - set to zero for no distance weighting

    norm_kernel:       set to 1 to normalise kernel weights
                       
    Returns:
    
    A df of all lagged columns from input df
    
    '''

    arg_string=str(kernel_inner)+'_'+str(kernel_width)+'_'+str(kernel_power)+'_'+str(norm_kernel)

    df=df.fillna(0.0)
    if not(df.index.is_monotonic):
        df=df.sort_index()
 
    weights=build_kernel_weights(kernel_inner,kernel_width,kernel_power,norm_kernel)
     
    times,time_to_index,index_to_time=utilities._map_times(df)

    features=utilities._map_features(df)

    pgids,pgid_to_longlat,longlat_to_pgid,pgid_to_index,index_to_pgid,ncells,power=\
                                                          utilities._map_pgids_2d(df)
    
    tensor4d=utilities._build_4d_tensor(
                                       df,
                                       pgids,
                                       pgid_to_index,
                                       times,
                                       time_to_index,
                                       ncells,
                                       ncells,
                                       pgid_to_longlat,
                                       features,
                                       use_stride_tricks
                                       )

    splags=get_splags(tensor4d,ncells,ncells,times,features,time_to_index,weights)

    df_splags=splags_to_df(
                           df,
                           splags,
                           times,
                           time_to_index,
                           pgids,
                           features,
                           ncells,
                           ncells,
                           longlat_to_pgid,
                           arg_string
                           )

    return df_splags

def build_kernel_weights(kernel_inner,kernel_width,kernel_power,norm_kernel):
    
    kernel_inner=int(kernel_inner)
    kernel_width=int(kernel_width)
    
    kernel_size=int(2*(kernel_inner+kernel_width))-1
    weights=np.ones((kernel_size,kernel_size))
    
    kernel_centre=int((kernel_size+1)/2)-1
    
    for ix in range(kernel_size):
        dx=ix-kernel_centre
        for iy in range(kernel_size):
            dy=iy-kernel_centre
            if ((abs(dx)<kernel_inner) and (abs(dy)<kernel_inner)):
               weights[ix,iy]=0.0
            else:
               rxy=np.sqrt(dx*dx+dy*dy)
               weights[ix,iy]/=rxy**kernel_power
               
    if (norm_kernel):
       weights/=np.sum(weights)
       
    return weights
    
def get_splags(
    tensor4d,
    longrange,
    latrange,
    times,
    features,
    time_to_index,
    weights
):

    splags=np.empty((longrange,latrange,len(times),len(features)))
    
    # use scipy convolution function to do 2d convolution on tensor slices
    
    for time in times:
        tindex=time_to_index[time]
        
        for ifeature in range(len(features)):
            raw=tensor4d[:,:,tindex,ifeature]
            splags[:,:,tindex,ifeature]=ndimage.convolve(raw, weights, mode='constant', cval=0.0)

    return splags
    
def splags_to_df(df,
    splags,
    times,
    time_to_index,
    pgids,
    features,
    longrange,
    latrange,
    longlat_to_pgid,
    argstring
):

    final=np.empty((len(times)*len(pgids),len(features)))

    # convert 4d spatial lag tensor to flat table

    ipgid=0
    npg=len(pgids)
    pgids_for_index=[]

    for ilong in range(1,longrange+2):
        for ilat in range(1,latrange+2):
            try:
                pgid=longlat_to_pgid[(ilong,ilat)]
                pgids_for_index.append(pgid)
                for time in times:
                    tindex=time_to_index[time]
                    indx=tindex*npg+ipgid
                    final[indx,:]=splags[ilong,ilat,tindex,:]

                ipgid+=1
            except:
                pass
            
    index_names=df.index.names
        
    df_index=pd.MultiIndex.from_product([list(times),pgids_for_index],names=index_names)
        
    colnames=['splag_'+argstring+'_'+feature for feature in features]
    df_splags=pd.DataFrame(data=final,columns=colnames,index=df_index)

    df_splags=df_splags.sort_index()

    return df_splags
    
    




""" Transform library.

Many functions assume data in the form of a pandas series or dataframe
indexed by timevar as level 0 and groupvar as level 1.

Originally written by Frederick Hoyles and Remco B. Jansen for Views 2
"""

import numpy as np  # type: ignore
import pandas as pd  # type: ignore

def delta(s: pd.Series, time: int = 1) -> pd.Series:
    """ Return the time-delta of s """

    return s - tlag(s, time=time)

def greater_or_equal(s: pd.Series, value: float) -> pd.Series:
    """ 1 if s >= value, else 0 """

    mask = s >= value
    y = mask.astype(int)

    return y

def smaller_or_equal(s: pd.Series, value: float) -> pd.Series:
    """ 1 if s >= value, else 0 """

    mask = s <= value
    y = mask.astype(int)

    return y

def in_range(s: pd.Series, low: float, high: float) -> pd.Series:
    """ 1 if low <= s <= high else 0 """

    y_high = smaller_or_equal(s, high)
    y_low = greater_or_equal(s, low)
    y = y_high + y_low - 1

    return y

def tlag(s: pd.Series, time: int) -> pd.Series:
    """ Time lag """
    if time < 1:
        msg = f"Time below 1 passed to tlag: {time} \n"
        msg += "Call tlead() instead \n"
        raise RuntimeError(msg)

    return s.groupby(level=1).shift(time)


def tlead(s: pd.Series, time: int) -> pd.Series:
    """ Time lead """
    if time < 1:
        msg = f"Time below 1 passed to tlead: {time} \n"
        msg += "Call tlag() instead \n"
        raise RuntimeError(msg)

    return s.groupby(level=1).shift(-time)

def moving_average(s: pd.Series, time: int) -> pd.Series:
    """ Moving average """
    if time < 1:
        msg = f"Time below 1 passed to ma: {time} \n"
        raise RuntimeError(msg)

    # Groupby groupvar
    y = s.groupby(level=1)
    # Divide into rolling time window of size time
    # min_periods=0 lets the window grow with available data
    # and prevent the function from inducing missingness
    y = y.rolling(time, min_periods=0)
    # Compute the mean
    y = y.mean()
    # groupby and rolling do stuff to indices, return to original form
    y = y.reset_index(level=0, drop=True).sort_index()
    return y

def moving_sum(s: pd.Series, time: int) -> pd.Series:
    """ Moving sum """
    if time < 1:
        msg = f"Time below 1 passed to ms: {time} \n"
        raise RuntimeError(msg)

    # Groupby groupvar
    y = s.groupby(level=1)
    # Divide into rolling time window of size time
    # min_periods=0 lets the window grow with available data
    # and prevent the function from inducing missingness
    y = y.rolling(time, min_periods=0)
    # Compute the sum
    y = y.sum()
    # groupby and rolling do stuff to indices, return to original form
    y = y.reset_index(level=0, drop=True).sort_index()
    return y


def cweq(s: pd.Series, value: float, seed=None) -> pd.Series:
    """Count while s equals value

    @TODO: Seed from series (series of seeds per groupvar?)

    """

    def set_seed(count, s, seed, mask):
        """Set count=seed in first time if mask was True there

        Example: We want time since conflict, which is time in peace.
        So we want count_while(conflict == 0).
        If our conflict series starts at 0 we might assume some longer
        previous history of peace.
        As the time count is summed cumulatively we can "seed" this
        counting sum with a starting value.

        This seed is therefre insterted into the first time period
        of the count IF the country is in peace at that time.
        Being in peace means the count is True, or ==1 as we
        already cast the masks T/F to the counters 1/0.

        """
        ix_timevar = s.index.get_level_values(0)
        first_time = ix_timevar == min(ix_timevar)
        mask_true = mask == 1
        first_time_where_mask_true = first_time & mask_true
        count.loc[first_time_where_mask_true] = seed
        return count

    # Drop NaN's
    s = s.dropna()

    # Boolean mask of where our condition (s==value) is True
    mask = s == value

    # This is a tricky one, print it out if its confusing.
    # Values of block_grouper are incremented when mask is NOT true.
    # This creates values that are constant (not incrementing) through a
    # consecutive spell of mask being True.
    # Grouping by this var thus lets the count.cumsum() restart for
    # each group of consecutive rows where mask is True and stay at
    # zero for the rows where block_grouper keeps incrementing,
    # which are the rows where mask is not met.
    # Note that mask is True when the criteria is fullfilled
    # Basically lets us assign a grouping id to each consecutive
    # spell of our condition being True.
    block_grouper = (~mask).groupby(level=1).cumsum()

    # Our mask becomes the basis for the count by casting it to int
    count = mask.astype(int)

    if seed:
        count = set_seed(count, s, seed, mask)

    # Get the groupvar-level index to group by
    ix_groupvar = s.index.get_level_values(1)

    # The time elapsed while condition is true
    y = count.groupby([block_grouper, ix_groupvar]).cumsum()
    y = y.astype(int)

    return y

def time_since(s, value=0, seed=None) -> pd.Series:
    """time since event in s where event is value other than 0.

    In order to compute a variable like "time since previous conflict
    event" we must apply a timelag to cweq() to get a series because
    for fitting a simultanous model we do not want the counter to be
    simultaneous to the event.

    Consider the data:

    event  : 0, 0, 1, 1, 0, 0 # Event
    cweq_0 : 1, 2, 0, 0, 1, 2 # count event while equals zero
    tisiev : ., 1, 2, 0, 0, 1 # time since event

    Fitting a model like "event ~ cweq0" makes no sense as cweq0 is
    always 0 if event=1.
    A model like "event ~ tsnp" makes more sense.
    We must apply a time lag to event before computing the counter to
    see how long time has elapsed since the previous event.

    Of course this isn't necessary for OSA modelling where all the
    rhs variables are time-lagged anyway but this is useful for
    dynamic simulation where X and predicted y are simulatenous.

    """

    return cweq(s=tlag(s=s, time=1), value=value, seed=seed)

def decay(s: pd.Series, halflife: float) -> pd.Series:
    """Decay function

    See half-life formulation at
    https://en.wikipedia.org/wiki/Exponential_decay
    """

    return 2 ** ((-1 * s) / halflife)

def mean(s: pd.Series) -> pd.Series:
    """ Per-groupvar arithmetic mean """

    return s.groupby(level=1).transform("mean")

def ln(s: pd.Series) -> pd.Series:
    """ Natural log of s+1 """
    return np.log1p(s)

def demean(s: pd.Series) -> pd.Series:
    """ demean, s = s - mean_group(s) """
    s_mean = s.groupby(level=1).transform("mean")
    return s - s_mean

def rollmax(s: pd.Series, window: int) -> pd.Series:
    """ Rolling max """
    # See https://github.com/pandas-dev/pandas/issues/14013
    y = s.groupby(level=1).apply(
        lambda x: x.rolling(window=window, min_periods=0).max()
    )

    return y

def onset_possible(s: pd.Series, window: int) -> pd.Series:
    """Onset possible if no event occured in the preceeding window times """
    # fillna() is so that the first t in a group is always a possible onset
    return (~rollmax(tlag(s, 1).fillna(0), window).astype(bool)).astype(int)

def onset(s: pd.Series, window: int) -> pd.Series:
    """Compute onset

    A row is defined as an onset if
    * onset is possible
    * s is greater than 0
    """
    s_onset_possible = (
        onset_possible(s, window).astype(bool) & s.astype(bool)
    ).astype(int)
    return s_onset_possible

def tick_time_since(s_event: pd.Series, s_time_since: pd.Series) -> np.ndarray:
    """Special time since ticker

    Special case time_sine counter.
    In dynasim we need a fast update of time_since variables.
    The normal time_since(s) is very fast for a single pass but for updates
    at each t it is too slow.
    This function uses previously computed time since column and the event
    column to speed up computation.
    Use only in loops over t and pass in only two times at a time, t and t-1
    like so:
    for t in times:
        df.loc["tick_time_since"] = tick_time_since(
            s_event=df.loc[t-1:t, "event"],
            s_time_since=df.loc[t-1:t, "time_since_event"]
        )

    The first t must be at least the THIRD time in the available times
    so that the original "time_since_event" col is defined properly, which
    happens in the second t as it uses a lag (missing in first t)
    and can be lagged by 1.


    """
    # Event from previous time
    event_lag = tlag(s_event, 1).dropna()
    # Time since count from previous time
    ts_lag = tlag(s_time_since, 1).dropna()
    # If no event in previous time, increment the time_since counter
    # If there was an event in previous t, return 0
    # Numpy+pandas magic takes care of the broadcasting when
    # assigning back in with df.loc[t, "tick"]
    # return np.where(~event_lag.astype(bool), ts_lag + 1, 0)
    return np.where(event_lag.astype(bool), 0, ts_lag + 1)



""" Missing data filling functionality 

Transformations for handling missing values, such as simple replacements, and
more advanced extrapolations.

"""

from typing import Any, List, Optional, Literal
import multiprocessing as mp

import numpy as np  # type: ignore
import pandas as pd  # type: ignore

from sklearn.experimental import enable_iterative_imputer  # type: ignore # noqa: F401, E501 # pylint: disable=unused-import
from sklearn.impute import IterativeImputer  # type: ignore
from sklearn.linear_model import BayesianRidge  # type: ignore

def replace_na(df: pd.DataFrame, replacement = 0):
    return df.replace(np.nan,replacement)

def list_totally_missing(df: pd.DataFrame) -> List[str]:
    """ Get a list of columns for which all values are missing """

    cols = []
    for col in df:
        if df[col].isnull().mean() == 1.0:
            cols.append(col)

    return cols



def fill_groups_with_time_means(df: pd.DataFrame) -> pd.DataFrame:
    """ Fill completely missing groups with time means """

    # Only fill numeric cols
    cols = list(df.select_dtypes(include=[np.number]).columns.values)
    for _, g_df in df.groupby(level=1):
        # If missing everything from a group
        if g_df.isnull().all().all():
            # Get the times for this group
            times_group = g_df.index.get_level_values(0)
            # Fill all columns with the time mean
            df.loc[g_df.index, cols] = (
                df.loc[times_group, cols].groupby(level=0).mean().values
            )
    return df



def fill_with_group_and_global_means(df: pd.DataFrame) -> pd.DataFrame:
    """ Impute missing values to group-level or global means. """

    for col in df.columns:
        # impute with group level mean
        df[col].fillna(
            df.groupby(level=1)[col].transform("mean"), inplace=True
        )
        # fill remaining NaN with df level mean
        df[col].fillna(df[col].mean(), inplace=True)

    return df



def extrapolate(
    df: pd.DataFrame,
    limit_direction: str = "both",
    limit_area: Optional[str] = None,
) -> pd.DataFrame:
    """ Interpolate and extrapolate """
    return (
        df.sort_index()
        .groupby(level=1)
        .apply(
            lambda group: group.interpolate(
                limit_direction=limit_direction, limit_area=limit_area
            )
        )
    )

def _fill_by_group(
    group: Any,
    limit_direction: Literal["forward", "backward", "both"],
    limit_area: Optional[Literal["inside", "outside"]],
) -> Any:
    # Get the outer boundaries of the group data.
    first_id = group.first_valid_index()
    last_id = group.last_valid_index()
    # Fill group according to set params.
    if limit_area is not None:
        # Assume forward if default "both" is passed with area "inside".
        if limit_area == "inside" and limit_direction != "backward":
            group[first_id:last_id] = group[first_id:last_id].ffill()
        if limit_area == "inside" and limit_direction == "backward":
            group[first_id:last_id] = group[first_id:last_id].bfill()
        if limit_area == "outside":

            id_min, id_max = group.index.min(), group.index.max()
            group[id_min:first_id] = group[id_min:first_id].bfill()
            group[last_id:id_max] = group[last_id:id_max].ffill()
    elif limit_direction == "forward":
        group = group.ffill()
    elif limit_direction == "backward":
        group = group.bfill()
    else:
        group = group.ffill().bfill()

    return group



def fill(
    s: pd.Series,
    limit_direction: Literal["forward", "backward", "both"] = "both",
    limit_area: Optional[Literal["inside", "outside"]] = None,
) -> pd.Series:
    """ Fill column in dataframe with optional direction and area.

    Args:
        s: Pandas series to apply filling to.
        limit_direction: Direction in which to fill.
        limit_area: Area to fill. Default None refers to the entire series.
    """

    
    return (
        s.sort_index()
        .groupby(level=1)
        .apply(
            lambda group: _fill_by_group(
                group=group,
                limit_direction=limit_direction,
                limit_area=limit_area,
            ),
        )
    )



def _fill_iterative(
    df: pd.DataFrame,
    seed: int = 1,
    max_iter: int = 10,
    estimator: Any = BayesianRidge(),
):
    """ Gets a single imputation using IterativeImputer from sklearn.

    Uses BayesianRidge() from sklearn.

    Changed default of sample_posterior to True as we're doing
    multiple imputation.

    Clips imputed values to min-max of observed values to avoid
    brokenly large values. When imputation model doesn't converge
    nicely we otherwise end up with extreme values that are out of
    range of the float32 type used by model training, causing crashes.
    Consider this clipping a workaround until a more robust imputation
    strategy is in place.

    """
    # Only impute numberic cols
    cols_numeric = list(df.select_dtypes(include=[np.number]).columns.values)
    cols_not_numeric = [col for col in df.columns if col not in cols_numeric]

    # Get bounds so we can clip imputed values to not be outside
    # observed values
    observed_min = df[cols_numeric].min()
    observed_max = df[cols_numeric].max()

    df_imputed = df.loc[:, []].copy()
    for col in df:
        df_imputed[col] = np.nan

    df_imputed[cols_numeric] = IterativeImputer(
        random_state=seed, max_iter=max_iter, estimator=estimator
    ).fit_transform(df[cols_numeric])
    df_imputed[cols_not_numeric] = df[cols_not_numeric]

    # Clip imputed values to observed min-max range
    df_imputed[cols_numeric] = df_imputed[cols_numeric].clip(
        observed_min, observed_max, axis=1
    )

    return df_imputed


def impute_mice_generator(
    df, n_imp, estimator=None, parallel=False, n_jobs=mp.cpu_count()
):
    """ Impute df with MICE """

    if parallel:
        with mp.Pool(processes=n_jobs, maxtasksperchild=1) as pool:
            results = [
                pool.apply_async(_fill_iterative, (df, imp, 10, estimator,))
                for imp in range(n_imp)
            ]
            for result in results:
                yield result.get()

    else:

        for imp in range(n_imp):

            yield _fill_iterative(df, seed=imp, estimator=estimator)



import numpy as np
import pandas as pd
from scipy.spatial import cKDTree

import utilities


def get_spacetime_distances(df,return_values='distances',k=1,nu=1.0,power=0.0):
    '''
    
    get_spacetime_distances
    
    For every point in the supplied df, uses scipy.spatial cKDTree to find the nearest k
    past events (where an event is any non-zero value in the input df) and returns either 
    the mean spacetime distance to the k events, or the mean of 
    (size of event)/(spacetime distance)**power.
    
    Arguments:
    
    df:            input df
    
    return_values: choice of what to return. Allowed values:
     
                   distances - return mean spacetime distance to nearest k events     
    
                   weights   - return mean of (size of event)/(spacetime distance)**power
                   
    k:             number of nearest events to be averaged over
    
    nu:            weighting to be applied to time-component of distances, so that 
                   spacetime distance = sqrt(delta_latitude^2 + delta_longitude^2 +
                   nu^2*delta_t^2)
                   
    power:         power to which distance is raised when computing weights. Negative
                   values are automatically converted to positive values.
    
    '''

    if return_values not in ['distances','weights']:
        raise Exception("unknown return_values; ",return_values,
                        " - allowed choices: distances weights")
                        
    power=np.abs(power)

    times,time_to_index,index_to_time=utilities._map_times(df)

    pgids,pgid_to_longlat,longlat_to_pgid,pgid_to_index,index_to_pgid,ncells,pwr=\
                                                           utilities._map_pgids_2d(df)

    features=utilities._map_features(df)

    use_stride_tricks=True

    tensor=utilities._build_4d_tensor(
                                     df,
                                     pgids,
                                     pgid_to_index,
                                     times,
                                     time_to_index,
                                     ncells,
                                     ncells,
                                     pgid_to_longlat,
                                     features,
                                     use_stride_tricks
                                     )

    df_stdist=space_time_distances(
                                    df,
                                    tensor,
                                    times,
                                    pgids,
                                    features,
                                    longlat_to_pgid,
                                    time_to_index,
                                    ncells,
                                    return_values,
                                    k,
                                    nu,
                                    power
                                    )
                                    
    return df_stdist

def space_time_distances(
                          df,
                          tensor,
                          times,
                          pgids,
                          features,
                          longlat_to_pgid,
                          time_to_index,
                          ncells,
                          return_values,
                          k,
                          nu,
                          power
                          ):
                          
    '''
    
    space_time_distances
    
    Finds spacetime distances, scaling spatial distances to degrees (for continuity with
    previous versions of this function) and stretching the time axis by nu.
    
    If k>1, averaging is performed.
    
    '''                      

    PGID_TO_DEGREES=0.5

    final=np.empty((len(times)*len(pgids),len(features)))

    npg=len(pgids)
    pgids_for_index=[]


    for ilong in range(ncells):
        for ilat in range(ncells):
            if (ilong,ilat) in longlat_to_pgid.keys():
                pgid=longlat_to_pgid[(ilong,ilat)]
                pgids_for_index.append(pgid)

    for time in times[0:600]:
    
        tindex=time_to_index[time]
        ipgid=0

        points=np.array(np.where(tensor[:,:,:tindex+1,0]>0)).astype(float).T
        

        if len(points)==0:
            pass
        else: 
            points[:,0]*=PGID_TO_DEGREES
            points[:,1]*=PGID_TO_DEGREES
            points[:,2]*=nu
            btree = cKDTree(data=points,leafsize=20)
                
        for ilong in range(ncells):
            for ilat in range(ncells):

                try:   
   
                    pgid=longlat_to_pgid[(ilong,ilat)]
                  
                    if len(points)==0:

                        feature=999.
 
                    else:

                        sptime_dists,ipoints=btree.query([ilong*PGID_TO_DEGREES,ilat*PGID_TO_DEGREES,nu*tindex],k)

                        if k==1:
                            ipoints=[ipoints,]
                            sptime_dists=[sptime_dists,]
                                
                        if return_values=='distances':
                                
                            if k==1:
                                feature=sptime_dists[0]
                            else:   
                                feature=np.mean(sptime_dists)
                        
                        elif return_values=='weights':

                            featurei=np.zeros(k)
                           
                            for i,ipoint in enumerate(ipoints):

                                coords=points[ipoint]
                                
                                ilongp=int(coords[0]/PGID_TO_DEGREES)
                                ilatp=int(coords[1]/PGID_TO_DEGREES)
                                itimep=int(coords[2]/nu)

                                if sptime_dists[i]==0.0:
                                    featurei[i]=tensor[ilongp,ilatp,itimep,0]
                                else:
                                    featurei[i]=tensor[ilongp,ilatp,itimep,0]/(sptime_dists[i]**(power))
                        
                            feature=np.mean(featurei)
                
                    indx=tindex*npg+ipgid
                    
                    final[indx,0]=feature

                    ipgid+=1
                except:
                    pass

    index_names=df.index.names
        
    df_index=pd.MultiIndex.from_product([times, pgids_for_index],names=index_names) 
    
    if return_values=='distances':
        colnames=['st_distances_nu_'+str(nu)+'_'+feature for feature in features]
    else:
        colnames=['st_weights_nu_'+str(nu)+'_k_'+str(k)+'_pwr_'+str(power)+'_'+feature for feature in features]
       
    df_stdist=pd.DataFrame(data=final,columns=colnames,index=df_index)

    df_stdist=df_stdist.sort_index()

    return df_stdist



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['views_transformation_library']

package_data = \
{'': ['*']}

install_requires = \
['ingester3>=0.6.0',
 'pandas>=1.2.3,<2.0.0',
 'scikit_learn==1.0.1',
 'scipy>=1.6.2,<2.0.0',
 'stepshift>=1.2.0',
 'xarray>=0.19.0']

setup_kwargs = {
    'name': 'views-transformation-library',
    'version': '2.3.0',
    'description': '',
    'long_description': None,
    'author': 'peder2911',
    'author_email': 'pglandsverk@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<3.10',
}


setup(**setup_kwargs)
